--- /dev/null	Wed Oct 28 00:34:10 2015
+++ new/usr/src/lib/libc/port/gen/pty.c	Wed Oct 28 00:34:10 2015
@@ -0,0 +1,155 @@
+/*
+ * Copyright 2015 Nexenta Systems, Inc. All rights reserved.
+ * Copyright (c) 1990, 1993, 1994
+ *	The Regents of the University of California.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 4. Neither the name of the University nor the names of its contributors
+ *    may be used to endorse or promote products derived from this software
+ *    without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
+
+
+#include <sys/types.h>
+#include <stropts.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <grp.h>
+#include <stdlib.h>
+#include <string.h>
+#include <termios.h>
+#include <unistd.h>
+#include <limits.h>
+
+int
+openpty(int *amaster, int *aslave, char *name, const struct termios *termp,
+    const struct winsize *winp)
+{
+	const char *slavename;
+	int master, slave;
+	int err;
+
+	if ((master = open("/dev/ptmx", O_RDWR|O_NOCTTY)) == -1)
+		return (-1);
+
+	if (grantpt(master) == -1)
+		goto bad;
+
+	if (unlockpt(master) == -1)
+		goto bad;
+
+	slavename = ptsname(master);
+	if (slavename == NULL) {
+		goto bad;
+	}
+
+	if ((slave = open(slavename, O_RDWR)) == -1)
+		goto bad;
+
+	if (ioctl(slave, I_PUSH, "ptem") < 0 ||
+	    ioctl(slave, I_PUSH, "ldterm") < 0 ||
+	    ioctl(slave, I_PUSH, "ttcompat") < 0) {
+		goto bad;
+	}
+
+	if (termp != NULL) {
+		if (tcsetattr(slave, TCSAFLUSH, termp) != 0) {
+			goto bad;
+		}
+	}
+	if (winp != NULL) {
+		if (ioctl(slave, TIOCSWINSZ, winp) != 0)
+			goto bad;
+	}
+
+	*amaster = master;
+	*aslave = slave;
+
+	if (name != NULL)
+		(void) strncpy(name, slavename, TTYNAME_MAX);
+	return (0);
+
+bad:
+	err = errno;
+	if (master != -1)
+		(void) close(master);
+	if (slave != -1)
+		(void) close(slave);
+	errno = err;  /* restore the original value of errno */
+	return (-1);
+}
+
+int
+login_tty(int fd)
+{
+
+	if (setsid() < 0) {
+		return (-1);
+	}
+
+	if (ioctl(fd, TIOCSCTTY, NULL) < 0)
+		return (-1);
+	if (dup2(fd, 0) < 0 ||
+	    dup2(fd, 1) < 0 ||
+	    dup2(fd, 2) < 0) {
+		return (-1);
+	}
+	if (fd > 2 && close(fd) < 0)
+		return (-1);
+	return (0);
+}
+
+int
+forkpty(int *amaster, char *name, const struct termios *termp,
+    const struct winsize *winp)
+{
+	int master, slave, pid;
+	int err;
+
+	if (openpty(&master, &slave, name, termp, winp) == -1)
+		return (-1);
+	switch (pid = fork()) {
+	case -1:
+		err = errno;
+		(void) close(master);
+		(void) close(slave);
+		errno = err;  /* restore the original value of errno */
+		return (-1);
+		}
+	case 0:
+		/*
+		 * child
+		 */
+		(void) close(master);
+		if (login_tty(slave) < 0) {
+			return (-1);
+		}
+		return (0);
+	}
+	/*
+	 * parent
+	 */
+	*amaster = master;
+	(void) close(slave);
+
+	return (pid);
+}
